@@ -58,6 +58,12 @@ typedef void (* GrdRdpFrameReadyCallback) (GrdRdpPipeWireStream *stream,
                                            gboolean              success,
                                            gpointer              user_data);
 
+typedef struct
+{
+  GMutex buffer_mutex;
+  gboolean is_locked;
+} BufferContext;
+
 struct _GrdRdpFrame
 {
   gatomicrefcount refcount;
@@ -129,6 +135,8 @@ struct _GrdRdpPipeWireStream
   struct pw_stream *pipewire_stream;
   struct spa_hook pipewire_stream_listener;
 
+  GHashTable *pipewire_buffers;
+
   uint32_t src_node_id;
 
   struct spa_video_info_raw spa_format;
@@ -141,6 +149,57 @@ static void grd_rdp_frame_unref (GrdRdpFrame *frame);
 
 G_DEFINE_AUTOPTR_CLEANUP_FUNC (GrdRdpFrame, grd_rdp_frame_unref)
 
+static BufferContext *
+buffer_context_new (void)
+{
+  BufferContext *buffer_context;
+
+  buffer_context = g_new0 (BufferContext, 1);
+  g_mutex_init (&buffer_context->buffer_mutex);
+
+  return buffer_context;
+}
+
+static void
+buffer_context_free (BufferContext *buffer_context)
+{
+  g_mutex_clear (&buffer_context->buffer_mutex);
+
+  g_free (buffer_context);
+}
+
+static void
+acquire_pipewire_buffer_lock (GrdRdpPipeWireStream *stream,
+                              struct pw_buffer     *buffer)
+{
+  BufferContext *buffer_context = NULL;
+
+  if (!g_hash_table_lookup_extended (stream->pipewire_buffers, buffer,
+                                     NULL, (gpointer *) &buffer_context))
+    g_assert_not_reached ();
+
+  g_mutex_lock (&buffer_context->buffer_mutex);
+  g_assert (!buffer_context->is_locked);
+  buffer_context->is_locked = TRUE;
+}
+
+static void
+maybe_release_pipewire_buffer_lock (GrdRdpPipeWireStream *stream,
+                                    struct pw_buffer     *buffer)
+{
+  BufferContext *buffer_context = NULL;
+
+  if (!g_hash_table_lookup_extended (stream->pipewire_buffers, buffer,
+                                     NULL, (gpointer *) &buffer_context))
+    g_assert_not_reached ();
+
+  if (!buffer_context->is_locked)
+    return;
+
+  buffer_context->is_locked = FALSE;
+  g_mutex_unlock (&buffer_context->buffer_mutex);
+}
+
 static GrdRdpFrame *
 grd_rdp_frame_new (GrdRdpPipeWireStream     *stream,
                    GrdRdpFrameReadyCallback  callback,
@@ -472,6 +531,33 @@ on_stream_param_changed (void                 *user_data,
                            params, G_N_ELEMENTS (params));
 }
 
+static void
+on_stream_add_buffer (void             *user_data,
+                      struct pw_buffer *buffer)
+{
+  GrdRdpPipeWireStream *stream = user_data;
+
+  g_hash_table_insert (stream->pipewire_buffers, buffer, buffer_context_new ());
+}
+
+static void
+on_stream_remove_buffer (void             *user_data,
+                         struct pw_buffer *buffer)
+{
+  GrdRdpPipeWireStream *stream = user_data;
+  BufferContext *buffer_context = NULL;
+
+  if (!g_hash_table_lookup_extended (stream->pipewire_buffers, buffer,
+                                     NULL, (gpointer *) &buffer_context))
+    g_assert_not_reached ();
+
+  /* Ensure buffer is not locked any more */
+  g_mutex_lock (&buffer_context->buffer_mutex);
+  g_mutex_unlock (&buffer_context->buffer_mutex);
+
+  g_hash_table_remove (stream->pipewire_buffers, buffer);
+}
+
 static void
 process_mouse_pointer_bitmap (GrdRdpPipeWireStream *stream,
                               struct spa_buffer    *buffer,
@@ -631,9 +717,10 @@ on_framebuffer_ready (gboolean success,
 
 static void
 process_frame_data (GrdRdpPipeWireStream *stream,
-                    struct spa_buffer    *buffer,
+                    struct pw_buffer     *pw_buffer,
                     GrdRdpFrame          *frame)
 {
+  struct spa_buffer *buffer = pw_buffer->buffer;
   GrdRdpFrameReadyCallback callback = frame->callback;
   gpointer user_data = frame->callback_user_data;
   uint32_t drm_format;
@@ -728,6 +815,7 @@ process_frame_data (GrdRdpPipeWireStream *stream,
       realize_buffer_data->hwaccel_nvidia = hwaccel_nvidia;
       realize_buffer_data->rdp_buffer = rdp_buffer;
 
+      acquire_pipewire_buffer_lock (stream, pw_buffer);
       grd_egl_thread_upload (egl_thread,
                              pbo,
                              height,
@@ -809,6 +897,7 @@ process_frame_data (GrdRdpPipeWireStream *stream,
           import_buffer_data->rdp_buffer = rdp_buffer;
         }
 
+      acquire_pipewire_buffer_lock (stream, pw_buffer);
       grd_egl_thread_download (egl_thread,
                                rdp_buffer->pbo,
                                height,
@@ -963,6 +1052,7 @@ out:
     pw_stream_queue_buffer (stream->pipewire_stream, buffer);
 
   g_source_set_ready_time (stream->render_source, 0);
+  maybe_release_pipewire_buffer_lock (stream, buffer);
 
   g_clear_pointer (&frame, grd_rdp_frame_unref);
 }
@@ -1021,7 +1111,7 @@ on_stream_process (void *user_data)
       return;
     }
 
-  process_frame_data (stream, last_frame_buffer->buffer,
+  process_frame_data (stream, last_frame_buffer,
                       g_steal_pointer (&frame));
 }
 
@@ -1029,6 +1119,8 @@ static const struct pw_stream_events stream_events = {
   PW_VERSION_STREAM_EVENTS,
   .state_changed = on_stream_state_changed,
   .param_changed = on_stream_param_changed,
+  .add_buffer = on_stream_add_buffer,
+  .remove_buffer = on_stream_remove_buffer,
   .process = on_stream_process,
 };
 
@@ -1316,12 +1408,18 @@ grd_rdp_pipewire_stream_finalize (GObject *object)
 
   g_mutex_clear (&stream->frame_mutex);
 
+  g_clear_pointer (&stream->pipewire_buffers, g_hash_table_unref);
+
   G_OBJECT_CLASS (grd_rdp_pipewire_stream_parent_class)->finalize (object);
 }
 
 static void
 grd_rdp_pipewire_stream_init (GrdRdpPipeWireStream *stream)
 {
+  stream->pipewire_buffers =
+    g_hash_table_new_full (NULL, NULL,
+                           NULL, (GDestroyNotify) buffer_context_free);
+
   g_mutex_init (&stream->frame_mutex);
 }
 
